from contextlib import contextmanager
from datetime import datetime
from time import time
import json
import os

from django.core import serializers
from django.conf import settings
if not hasattr(settings, 'SERIALIZATION_MODULES'):
    settings.SERIALIZATION_MODULES = {}
if 'django_snailtracker.serializer' not in settings.SERIALIZATION_MODULES:
    settings.SERIALIZATION_MODULES['django_snailtracker.serializer'] = 'django_snailtracker.serializer'

SNAILTRACKER_TMP_PATH = getattr(settings, 'SNAILTRACKER_TMP_PATH', '/tmp/')
SNAILTRACKER_TIMEOUT_SECONDS = getattr(settings, 'SNAILTRACKER_TIMEOUT_SECONDS', 5)


def snailtracker_enabled():
    return getattr(settings, 'SNAILTRACKER_ENABLED', False)


def make_model_snapshot(instance):
    object_as_dict = json.loads(
            serializers.serialize(
                'django_snailtracker.serializer', [instance]))[0]
    if hasattr(instance, 'snailtracker_exclude_fields'):
        for field in instance.snailtracker_exclude_fields:
            try:
                del object_as_dict['fields'][field]
            except KeyError:
                pass
    return object_as_dict


def dict_diff(d1, d2):
    c_list = list()
    for k, v in d1.iteritems():
        if unicode(d2[k]) != unicode(v):
            c_list.append(k)
    return c_list


def dict_diff_with_values(dict1, dict2):
    the_diff = {}
    for k, v in dict1.iteritems():
        if unicode(dict2[k]) != unicode(v):
            the_diff[k] = {'old': dict1[k], 'new': dict2[k]}
    return the_diff


def diff_from_action(action, values=True):
    if action.post_init_snapshot and action.post_save_snapshot:
        if not values:
            return dict_diff(
                    action.init_snapshot_to_python()['fields'],
                    action.save_snapshot_to_python()['fields'])
        else:
            return dict_diff_with_values(
                    action.init_snapshot_to_python()['fields'],
                    action.save_snapshot_to_python()['fields'])


@contextmanager
def mutex_lock(key):
    mutex_key = '%ssnailtracker.%s.lock' % (SNAILTRACKER_TMP_PATH, key)
    if os.path.exists(mutex_key):
        with open(mutex_key, 'r') as tmp_f:
            file_data = tmp_f.read()
            if file_data:
                was_made = datetime.fromtimestamp(float(file_data))
                if (datetime.now() - was_made).seconds < SNAILTRACKER_TIMEOUT_SECONDS:
                    raise SnailtrackerMutexLockedError("'%s' is already locked" % (key,))
    with open(mutex_key, 'w') as tmp_f:
        tmp_f.write(unicode(time()))
    try:
        yield True
    finally:
        if os.path.exists(mutex_key):
            os.remove(mutex_key)


class SnailtrackerMutexLockedError(Exception):
    pass



ACTION_TYPE_INSERT = 1
ACTION_TYPE_UPDATE = 2
ACTION_TYPE_DELETE = 3

ACTION_TYPE_CHOICES = (
    (ACTION_TYPE_INSERT, 'insert'),
    (ACTION_TYPE_UPDATE, 'update'),
    (ACTION_TYPE_DELETE, 'delete'),
)



from django.core.paginator import Paginator, InvalidPage, EmptyPage

from django_snailtracker.models import Snailtrack, Action

def index(request):
    return {}

def table_index(request, table):
    st = Snailtrack(client=redis_factory())
    st.table = table
    pks = list(st.get_pks_for_table())
    paginator = Paginator(pks, 10)
    try:
        page = int(request.GET.get('page', '1'))
    except ValueError:
        page = 1
    try:
        pks = paginator.page(page)
    except (EmptyPage, InvalidPage):
        pks = paginator.page(paginator.num_pages)
    return {'table': table, 'pks': pks}

def view(request, table=None, pk=None):
    st = Snailtrack(client=redis_factory())
    try:
        st.get(table=table, primary_key=pk)
    except:
        st = None
    return {'snailtrack':st}



from django.core.serializers.json import Serializer as JSONSerializer
from django.core.serializers.python import Deserializer

class Serializer(JSONSerializer):
    """
    Convert a queryset to JSON with a minimum of extra db queries.
    """
    internal_use_only = False

    def handle_fk_field(self, obj, field):
        self._current[field.attname] = getattr(obj, field.attname)

    def handle_m2m_field(self, obj, field):
        # :MC: m2m relationships should maybe be tracked with parent/child relationships.
        # value = list(getattr(obj, field.name).values_list('pk', flat=True))
        # if len(value) <= 0:
        #     value = None
        # self._current[field.name] = value
        pass



from django.conf import settings
celery_enabled = False
if getattr(settings, 'SNAILTRACKER_OFFLOAD', False):
    try:
        from celery.task import task
        celery_enabled = True
    except ImportError:
        celery_enabled = False

from django_snailtracker.models import get_or_create_snailtrack


if celery_enabled:
    @task
    def offload_wrapper(instance, deleted=False):
        """
        This function if a celery task
        """
        get_or_create_snailtrack(instance=instance, deleted=deleted)
else:
    offload_wrapper = None



import logging

__version__ = '0.5.7'

logger = logging.getLogger(__name__)
logger.setLevel('DEBUG')


def autodiscover():
    """
    Auto-discover INSTALLED_APPS snailtracker.py modules and fail silently when
    not present. This forces an import on them to register any admin bits they
    may want.
    """

    from django.conf import settings
    from django.utils.importlib import import_module
    from django.utils.module_loading import module_has_submodule

    for app in settings.INSTALLED_APPS:
        mod = import_module(app)
        # Attempt to import the app's snailtracker module.
        try:
            import_module('%s.snailtracker' % app)
            logger.debug('found snailtracker module in %s' % app)
        except:
            # Decide whether to bubble up this error. If the app just
            # doesn't have an admin module, we can ignore the error
            # attempting to import it, otherwise we want it to bubble up.
            if module_has_submodule(mod, 'snailtracker'):
                raise



import logging

from django.db.models.signals import post_init, post_save, post_delete

from django_snailtracker.models import (snailtracker_post_init_hook,
        snailtracker_post_save_hook, snailtracker_post_delete_hook)
from django_snailtracker.sites import snailtracker_site


logger = logging.getLogger(__name__)


def register(obj_def):
    if obj_def._meta.db_table not in snailtracker_site.registry:
        logger.debug('Registering %s' % obj_def._meta.db_table)
        post_init.connect(snailtracker_post_init_hook, sender=obj_def,)
        post_save.connect(snailtracker_post_save_hook, sender=obj_def,)
        post_delete.connect(snailtracker_post_delete_hook, sender=obj_def,)
        snailtracker_site.registry[obj_def._meta.db_table] = True
    else:
        logger.debug('%s already registered' % obj_def._meta.db_table)



from setuptools import setup, find_packages

#version = __import__('django_snailtracker').__version__
version = '0.5.7'


setup(name='django-snailtracker',
      version=version,
      description="Track history on django models",
      long_description="""\
""",
      classifiers=[],  # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='django model-history-tracking',
      author='Kyle Terry',
      author_email='kyle@kyleterry.com',
      url='',
      license='New BSD License',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          # -*- Extra requirements: -*-
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )
